/*
 *   trimrom - A generic ROM dumps trimmer
 *
 *   Copyright (C) 2009 by
 *   eadmaster - <spamhereplease@fastwebnet.it> - <http://www.mantox.has.it>
 *
 *   'trimrom' is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *   See the GNU General Public License for more details.
 *   <http://www.gnu.org/licenses/gpl.txt> 
 */

/*
TODO: skip non-consecutive 0es and 1s.
TODO: check if output file already exists.
TODO: cmdline switches: -v N = verbosity, -q = quiet, -V = version, -i = inplace
TODO: inplace trimming using ~ftruncate
*/

#include <stdio.h>
#include <stdlib.h>

// define byte storage type
typedef unsigned char byte;
/*
#ifdef __ISOC99_SOURCE
    #include <stdint.h>
	typedef uint8_t byte;
#else
	typedef unsigned char byte;
#endif
*/

const char* program = "trimrom";
const char* version = "v0.4.1";
const char* author = "eadmaster";

void printVersion()
{
	printf("%s %s\n", program, version);
	puts("");
	printf("Copyright 2009 %s.\n", author);
	puts("This program is distributed without any warranty. USE IT AT YOUR OWN RISK!\n");
}

void printHelp()
{
	printf("usage: %s [input file] [output file]\n", program);
	puts("A generic ROM dumps trimmer. Trim trailing bytes set to 00 or FF.\n"
	     "Input file is left unmodified.\n"
	     "If no output file is specified only count trimmable bytes.\n"
	     "\n"
	     "Exit status is 0 if OK, 1 if I/O error.\n"
	     "\n"
	     "Report bugs to <spamhereplease@fastwebnet.it>.\n"
	);
}

void fatalError( char* errorDescription )
{
	printf("%s: %s.\n", program, errorDescription);
	exit(1);
}

void debugMessage( char* msg )
{
	printf("%s: %s\n", program, msg);
}




FILE* inputFile = NULL;
long int inputFileSize = 0;
//byte* inputFileReadBuffer = NULL;
FILE* outputFile = NULL;
long int outputFileSize = 0;


void readInputFileSize()
{
	fseek ( inputFile, 0, SEEK_END ); // seek to the end of file
	inputFileSize = ftell( inputFile ); // read inputFileSize
	rewind ( inputFile ); // -> fseek ( inputFile, 0, SEEK_??? );
	/* ALTERNATIVE (non necessita che il file sia aperto):
	struct stat file_status;
	stat(filename, &file_status);
	inputFileSize = file_status.st_size;
	*/
}

void countTrimmableBytes()
{	
	/* METHOD1: read bytes 1-by-1 */
	fseek ( inputFile, -1, SEEK_END ); // seek to the end of file
	
	outputFileSize = inputFileSize;
	byte currentValue = 0;
	while( outputFileSize > 0 )
	{
		if(
			fread( &currentValue, 1, 1, inputFile ) // read 1 byte and move 1 byte forward
		!= 1)
		{
			fatalError("I/O error reading input file");
		}
		
		if(currentValue!=0 && currentValue!=255)
		{
			break;
		}
		
		outputFileSize--;
		fseek( inputFile, -2, SEEK_CUR ); // go back 2 bytes
	}
	

	/* METHOD 2: use a read buffer (not good for BIG files!) */
	/*	
	inputFileReadBuffer = malloc( inputFileSize ); // allocate read buffer
	if( !inputFileReadBuffer )
		fatalError("unable to allocate read buffer (input file too big?)");
	
	if( fread( inputFileReadBuffer, 1, inputFileSize, inputFile ) < inputFileSize ) // read ALL input file into the buffer
		fatalError("I/O error reading input file");
	
	fclose ( inputFile );
	
	// count trimmable bytes
	outputFileSize = inputFileSize;
	
	while(outputFileSize>0)
	{
		byte currentValue = inputFileReadBuffer[outputFileSize];

		if(currentValue!=0 && currentValue!=255)
		{
			outputFileSize++;
			break;
		}
		
		outputFileSize--;
	}
	*/
}


void copyTrim()
{	
	/* METHOD1: write bytes 1-by-1 */
	rewind ( inputFile );
	byte currentValue = 0;
	while( outputFileSize > 0 )
	{
		if(
			fread( &currentValue, 1, 1, inputFile ) // read 1 byte from the input file
		!= 1)
		{
			fatalError("I/O error reading input file");
		}
		
		if(
			fwrite( &currentValue, 1, 1, outputFile ) // write 1 byte into the output file
		!= 1)
		{
			fatalError("I/O error writing output file");
		}
		
		outputFileSize--;
	}
	debugMessage("output file written corectly");
	
	/* METHOD2: use a read buffer (not good for BIG files!)
	if( fwrite( inputFileReadBuffer, 1, outputFileSize, outputFile ) < outputFileSize )
		fatalError("I/O error writing output file");
	else
		debugMessage("output file written corectly");
	*/

}


int main( int argc, char* argv[] )
{
	printVersion();
	
	if( argc <= 1 || *argv[1] == '-' )
	{
		printHelp();
		exit(0);
	}
	
	printf("%s: input file: \"%s\"\n", program, argv[1]);

	inputFile = fopen ( argv[1], "rb" ); // open the input file in read-only binary mode
	if( !inputFile )
		fatalError("unable to open input file");
	else
		debugMessage("input file opened for read only");

	readInputFileSize();
	countTrimmableBytes();
	
	printf("%s: trimable bytes: %u\n", program, inputFileSize - outputFileSize);

	if( argc <= 2 ) // if the output file is not specified then exit
	{
		printf("%s: no output file specified, exiting\n", program);
		fclose ( inputFile );
		exit(0);
	}
	
	printf("%s: output file: \"%s\"\n", program, argv[2]);
	
	outputFile = fopen ( argv[2], "wb" ); // open the output file in write-only binary mode
	if( !outputFile )
		fatalError("unable to open output file");

	copyTrim(); // write output file
	
	fclose( outputFile );
	fclose( inputFile );
}
